/*
 * system for managing buffered audio
 */

#include <string.h>
#include <libn64.h>
#include "audio.h"

// struct for a sound buffer to force it to be aligned to 8 bytes
union sb_align_u {
    unsigned long long longlongs[AUDIO_BUFFER_SAMPLES*2/4];
    short shorts[AUDIO_BUFFER_SAMPLES*2];
};

union sb_align_u sb[AUDIO_BUFFER_COUNT];

// number of done
unsigned int audio_buffers_deferred;
int audio_readbuf;
// where we're currently writing
int audio_writebuf;

#ifdef AUDIO_DEBUG
int audio_restarts;
#endif

#ifdef AUDIO_INTERRUPT_DRIVEN
void audio_AI_interrupt(void);
#endif
void audio_handle_deferred(int);

void audio_init(void)
{
    AI_set_frequency(AUDIO_FREQUENCY);

    audio_writebuf=0;
    audio_buffers_deferred=0;
#ifdef AUDIO_DEBUG
    audio_restarts=0;
#endif

    memset(sb,0,sizeof(sb));

#ifdef AUDIO_INTERRUPT_DRIVEN
    registerAIhandler(audio_AI_interrupt);
    set_AI_interrupt(1);
#endif
}

#ifdef AUDIO_INTERRUPT_DRIVEN
void audio_AI_interrupt(void) {
    audio_handle_deferred(0);
}
#endif

// with force set we are guaranteed to have < AUDIO_BUFFER_COUNT-2 deferred
// after this call
void audio_handle_deferred(int force)
{
    while (
            // if the AI is not full we might as well send buffers now
            (!AI_full() && audio_buffers_deferred > 0) ||
            // however even if the AI is full we may still need to send if we've
            // deferred too many
            (force && audio_buffers_deferred >= AUDIO_BUFFER_COUNT-2)
          ) {
#ifdef AUDIO_DEBUG
        if (!AI_active())
            audio_restarts++;
#endif
        AI_add_buffer(sb[audio_readbuf].shorts,AUDIO_BUFFER_SAMPLES);
        audio_readbuf++;
        if (audio_readbuf >= AUDIO_BUFFER_COUNT)
            audio_readbuf = 0;
        audio_buffers_deferred--;
    }
}

// Done writing to this buffer, let 'er rip
void audio_done(void)
{
#ifdef AUDIO_INTERRUPT_DRIVEN
    set_AI_interrupt(0);
#endif
    // make sure the AI can see the data
    data_cache_writeback(sb[audio_writebuf].longlongs,AUDIO_BUFFER_SAMPLES*4);

    // if this is the first on the deferred list, mark that this is the start
    if (audio_buffers_deferred == 0)
        audio_readbuf=audio_writebuf;

    // defer it
    audio_buffers_deferred++;

    // potentially undefer it
    audio_handle_deferred(0);

#ifdef AUDIO_INTERRUPT_DRIVEN
    set_AI_interrupt(1);
#endif

    // next time use a different buffer
    audio_writebuf++;
    if (audio_writebuf >= AUDIO_BUFFER_COUNT) audio_writebuf = 0;

}

// Get a buffer to write on.
short * audio_getbuf(void)
{
#ifdef AUDIO_INTERRUPT_DRIVEN
    set_AI_interrupt(0);
#endif
    // We can't get a buffer to write on if there are too many deferred,
    // as any buffer we get to write will be put on the deferred list at
    // least temorarily.
    if (audio_buffers_deferred >= AUDIO_BUFFER_COUNT-2)
        audio_handle_deferred(1);
#ifdef AUDIO_INTERRUPT_DRIVEN
    set_AI_interrupt(1);
#endif

    return sb[audio_writebuf].shorts;
}

int audio_getbufsamplecount(void)
{
    return AUDIO_BUFFER_SAMPLES;
}

